#include "storage.h"

#include <EEPROM.h>
#include <esp_attr.h>

constexpr uint8_t absoluteTemperatureIndicator = 0xFF;

RTC_DATA_ATTR uint8_t storageFlags = 0;

// Storage for temperature measurements in RTC memory that survives deep sleep
RTC_DATA_ATTR uint8_t data[rtcStorageSize];
RTC_DATA_ATTR uint16_t dataIndex = 0;
RTC_DATA_ATTR uint16_t numberOfMeasurements = 0;
RTC_DATA_ATTR uint32_t numberOfDiscardedMeasurements = 0;

// The index into EEPROM storage where the next data should be written
RTC_DATA_ATTR uint32_t eepromIndex = 0;
RTC_DATA_ATTR bool isFirstRunAfterPowerOn = true;

constexpr uint16_t eepromOffset = sizeof(uint16_t);
constexpr uint16_t eepromDataSize = eepromSize - eepromOffset;

// Keeps the last valid temperatures for each sensor
RTC_DATA_ATTR Temperature lastTemperatures[temperatureSensorCount];
RTC_DATA_ATTR uint32_t lastValidTemperatureTime[temperatureSensorCount];
RTC_DATA_ATTR uint8_t lastMeasurements[temperatureSensorCount] = {0, 0};

void setStorageFlag(StorageFlags flag) {
    storageFlags |= static_cast<uint8_t>(flag);
}

void clearStorageFlag(StorageFlags flag) {
    storageFlags &= ~static_cast<uint8_t>(flag);
}

bool hasStorageFlag(StorageFlags flag) {
    return storageFlags & static_cast<uint8_t>(flag);
}

bool hasInitializedEEPROM() {
    return hasStorageFlag(StorageFlags::HasInitializedEEPROM);
}

void didInitializeEEPROM() {
    setStorageFlag(StorageFlags::HasInitializedEEPROM);
}

void setupEEPROMIfNeeded() {
    if (hasInitializedEEPROM()) {
        return;
    }
    bool success = EEPROM.begin(eepromSize);
    if (!success) {
        Serial.println("Failed to set up EEPROM");
        setStorageFlag(StorageFlags::FailedToInitEEPROM);
    }
    Serial.print("EEPROM size: ");
    Serial.println(eepromSize);
    clearStorageFlag(StorageFlags::FailedToInitEEPROM);
    setStorageFlag(StorageFlags::HasInitializedEEPROM);
}

void resetLastMeasurements() {
    for (uint8_t index = 0; index < temperatureSensorCount; index += 1) {
        lastTemperatures[index].status == TemperatureStatus::sensorNotFound;
        lastTemperatures[index].value = 0;
        lastValidTemperatureTime[index] = 0;
        lastMeasurements[index] = 0;
    }
}

void storageConfigure(bool isFirstRun) {
    if (isFirstRun) {
        // Ensure that first values are stored
        resetLastMeasurements();
    }
    // Initialize EEPROM on every wake
    // setupEEPROMIfNeeded();
    //setStorageFlag(StorageFlags::HasInitializedEEPROM);
}

void moveDataToEEPROMIfNeeded() {
    if (dataIndex < rtcStorageSize) {
        return;
    }

    setupEEPROMIfNeeded();
    if (!hasInitializedEEPROM()) {
        return;
    }

    // Write until EEPROM is full
    uint16_t bytesRemaining = eepromDataSize - eepromIndex;
    uint16_t bytesToWrite = min(dataIndex, bytesRemaining);
    EEPROM.writeBytes(eepromIndex, data, bytesToWrite); // TODO: Check that result == bytesToWrite
    EEPROM.commit();

    // Move remaining data to front of array (if any)
    uint16_t bytesToKeep = dataIndex - bytesToWrite;
    memmove(data, data + bytesToWrite, bytesToKeep);
    dataIndex -= bytesToWrite;
}

void saveByteAtCurrentIndex(uint8_t byte) {
    if (dataIndex >= rtcStorageSize) {
        setStorageFlag(StorageFlags::RTCStorageOverflow);
        return;
    }
    data[dataIndex] = byte;
    dataIndex += 1;
    //moveDataToEEPROMIfNeeded();
}

uint8_t byteForAbsoluteTemperature(Temperature* temp) {
    switch (temp->status) {
        case TemperatureStatus::sensorError:
        return temperatureSensorFailure;
        case TemperatureStatus::sensorNotFound:
        return temperatureSensorNotAvailable;
        case TemperatureStatus::temperatureIsValid:
        break;
        default:
        return temperatureSensorNotAvailable;
    }
    // Convert to temperature range
    if (temp->value > 87500) {
        return temperatureMaximumValue;
    }
    if (temp->value < -39000) {
        return temperatureMinimumValue;
    }
    long converted = (temp->value - (temperatureShiftForStorage)) / 500;
    if (converted < temperatureMinimumValue) {
        return temperatureMinimumValue;
    }
    if (converted > temperatureMaximumValue) {
        return temperatureMaximumValue;
    }
    return converted;
}

bool needsAbsoluteTemperatureRecording(Temperature* temperatures) {
    if (numberOfMeasurements == 0) {
        return true;
    }
    for (uint8_t index = 0; index < temperatureSensorCount; index += 1) {
        if (temperatures[index].status != TemperatureStatus::temperatureIsValid) {
            // Error value can be encoded as differential (0x00)
            continue;
        }
        // Note: If previously only errors, then last value is 0. This also work for differential temperatures
        long diff = (lastTemperatures[index].value - temperatures[index].value) / 500 + 8;
        if (diff < 1 || diff > 15) {
            return true;
        }
    }
    return false;
}

void saveTemperatures(Temperature* temperatures) {
    if (needsAbsoluteTemperatureRecording(temperatures)) {
        // Write absolute temperatures
        saveByteAtCurrentIndex(absoluteTemperatureIndicator);
        for (uint8_t sensorIndex = 0; sensorIndex < temperatureSensorCount; sensorIndex += 1) {
            Temperature* temp = &temperatures[sensorIndex];
            uint8_t byte = byteForAbsoluteTemperature(temp);
            saveByteAtCurrentIndex(byte);
            lastMeasurements[sensorIndex] = byte;
            if (temp->status == TemperatureStatus::temperatureIsValid) {
                // Only update if temperature is valid
                lastTemperatures[sensorIndex] = *temp;
                lastValidTemperatureTime[sensorIndex] = time(NULL);
            }
        }
    } else {
        // Calculate temperature differences
        uint8_t valueToStore = 0;
        for (uint8_t index = 0; index < temperatureSensorCount; index += 1) {
            Temperature* temp = &temperatures[index];
            uint8_t diff = 0; // Indicate sensor error
            if (temp->status == TemperatureStatus::temperatureIsValid) {
                diff = (lastTemperatures[index].value - temp->value) / 500 + 8;
            }
            if (index % 2) {
                // Store second bits
                valueToStore |= diff; // Valid range already ensured
                saveByteAtCurrentIndex(valueToStore);
                valueToStore = 0;
            } else {
                // Store in first four bits
                valueToStore = (diff << 4);
            }
            lastMeasurements[index] = byteForAbsoluteTemperature(temp);

            if (temp->status == TemperatureStatus::temperatureIsValid) {
                // Only update if temperature is valid
                lastTemperatures[index].status = TemperatureStatus::temperatureIsValid;
                lastTemperatures[index].value = temp->value;
                lastValidTemperatureTime[index] = time(NULL);
            } else if (lastTemperatures[index].status != TemperatureStatus::temperatureIsValid) {
                lastTemperatures[index].value = 0;
                lastValidTemperatureTime[index] = 0;
            }
        }
        // Ensure storage with uneven number of sensors
        if (temperatureSensorCount % 2) {
            saveByteAtCurrentIndex(valueToStore);
        }
    }
    numberOfMeasurements += 1;
}

uint16_t getTotalNumberOfStoredBytes() {
    return dataIndex; // + eepromIndex;
}

uint16_t getNumberOfMeasurements() {
    return numberOfMeasurements;
}

uint32_t getTotalNumberOfMeasurements() {
    return numberOfMeasurements + numberOfDiscardedMeasurements;
}

uint8_t getLastTemperature(uint8_t sensorIndex) {
    return lastMeasurements[sensorIndex];
}

uint16_t getTimeSinceValidTemperature(uint8_t sensorIndex) {
    return time(NULL) - lastValidTemperatureTime[sensorIndex];
}

uint16_t getRecordedBytesAtOffset(uint8_t* buffer, uint16_t offset, uint16_t count) {
    // Copy remaining bytes from RTC memory
    uint16_t remainingBytes = dataIndex - offset;
    uint16_t bytesToCopy = min(count, remainingBytes);
    memcpy(buffer, data + offset, count);
    return bytesToCopy;
    /*
    // TODO: Check limits
    uint16_t eepromByteCount = eepromIndex;
    uint16_t endIndex = offset + count;
    uint16_t eepromStart = min(offset, eepromByteCount);
    uint16_t eepromEnd = min(endIndex, eepromByteCount);
    uint16_t bytesToCopyFromEEPROM = eepromEnd - eepromStart;
    if (bytesToCopyFromEEPROM > 0) {
        EEPROM.readBytes(eepromStart, buffer, bytesToCopyFromEEPROM); // TODO: Check bytes read
        // Reduce offset and count to point to RTC memory
        offset -= bytesToCopyFromEEPROM;
        count -= bytesToCopyFromEEPROM;
    }
    offset -= eepromByteCount;
    // Copy remaining bytes from RTC memory
    uint16_t bytesToCopyFromRTC = min(count, dataIndex);
    memcpy(buffer + bytesToCopyFromEEPROM, data + offset, bytesToCopyFromRTC);
    return bytesToCopyFromEEPROM + bytesToCopyFromRTC;
    */
}

void discardAllRecordedBytes() {
    eepromIndex = 0;
    dataIndex = 0;
    numberOfDiscardedMeasurements += numberOfMeasurements;
    numberOfMeasurements = 0;
}

uint8_t getStorageErrorFlags() {
    return storageFlags;
}